#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <gringo.h>

extern void exit_from_gringosh(void);


#define BUFFER_SIZE		255
#define	TOKBUFFER_SIZE	4096

// ===== Token Private vars =====
char strtok_buffer[TOKBUFFER_SIZE];
char *strtok_pointer;
int token_active = 0;	// Parse the token only if set to 1
int token_count = 0;	// Number of tokens inside of a line
int token_line  = 0;	// Number of processed lines

TObject *LastActiveWindow = NULL;

struct {
  int type;
  int x, y, w, h;
  char *ibuffer;
  char *lname;
  int flag;
} stoken;


void token_begin() {
  token_count = 0;
  token_line++;
  stoken.type = TYPE_UNKNOWN;
  stoken.x = 0;
  stoken.y = 0;
  stoken.w = 0;
  stoken.h = 0;
  if ( stoken.ibuffer != NULL ) free(stoken.ibuffer);
  stoken.ibuffer = NULL;
  if ( stoken.lname != NULL ) free(stoken.lname);
  stoken.lname = NULL;
  stoken.flag = 0;
}


int token_parse(char *s) {
  // Trim the token
  while ( s[0] == ' ' ) s++;
  while ( s[strlen(s)-1] == ' ' ) s[strlen(s)-1]=0;

  // Parse
  switch (token_count) {

    // First Token
    case 0: {
      if ( 0 == strcmp( "@WINDOW", s ) ) stoken.type = TYPE_WINDOW;
      else
      if ( 0 == strcmp( "@BUTTON", s ) ) stoken.type = TYPE_BUTTON;
      else
      if ( 0 == strcmp( "@GET", s ) ) stoken.type = TYPE_EDITBOX;
      else
      if ( 0 == strcmp( "@LABEL", s ) ) stoken.type = TYPE_LABEL;
      else
      return 1;
      //
      break;
    }

    // Seconth
    case 1: {
      switch (stoken.type) {
        // Expect Widget Name
	case TYPE_WINDOW: 
	case TYPE_BUTTON: 
	case TYPE_EDITBOX:
	{
	  //printf("Token at line %i is a name\n", token_line );
	  stoken.lname = strdup(s);
	  break;
	}
        // Expect "Y"
	case TYPE_LABEL: 
	{
	  //printf("Token at line %i is Y\n", token_line );
	  stoken.y = atoi(s);
	  break;
	}
	default: {
          printf( "GRINGO: Token %s at line %i, pos. %i is not required !\n", s, token_line, token_count );
	  return 1;
	  break;
        }
      }
      break;
    }

    // Thirth
    case 2: {
      switch (stoken.type) {
        // Expect "Y"
	case TYPE_WINDOW: 
	case TYPE_BUTTON: 
	case TYPE_EDITBOX:
	{
	  //printf("Token at line %i is Y\n", token_line );
	  stoken.y = atoi(s);
	  break;
	}
        // Expect "X"
	case TYPE_LABEL: 
	{
	  //printf("Token at line %i is X\n", token_line );
	  stoken.x = atoi(s);
	  break;
	}
	default: {
          printf( "GRINGO: Token %s at line %i, pos. %i is not required !\n", s, token_line, token_count );
	  return 1;
	  break;
        }
      }
      break;
    }

    // Fourth
    case 3: {
      switch (stoken.type) {
        // Expect "X"
	case TYPE_WINDOW: 
	case TYPE_BUTTON: 
	case TYPE_EDITBOX:
	{
	  //printf("Token at line %i is X\n", token_line );
	  stoken.x = atoi(s);
	  break;
	}
	// Expect W
	case TYPE_LABEL: 
	{
	  //printf("Token at line %i is W\n", token_line );
	  stoken.w = atoi(s);
	  break;
	}
	default: {
          printf( "GRINGO: Token %s at line %i, pos. %i is not required !\n", s, token_line, token_count );
	  return 1;
	  break;
        }
      }
      break;
    }

    // Fifth
    case 4: {
      switch (stoken.type) {
        // Expect "W"
	case TYPE_WINDOW: 
	case TYPE_BUTTON: 
	case TYPE_EDITBOX:
	{
	  //printf("Token at line %i is W\n", token_line );
	  stoken.w = atoi(s);
	  break;
	}
        // Expect "Title"
	case TYPE_LABEL:
	{
	  //printf("Token at line %i is Title\n", token_line );
	  stoken.ibuffer = strdup(s);
	  break;
	}
	default: {
          printf( "GRINGO: Token %s at line %i, pos. %i is not required !\n", s, token_line, token_count );
	  return 1;
	  break;
        }
      }
      break;
    }

    // Fifth
    case 5: {
      switch (stoken.type) {
        // Expect "H"
	case TYPE_WINDOW: 
	{
	  //printf("Token at line %i is H\n", token_line );
	  stoken.h = atoi(s);
	  break;
	}
        // Expect "Title"
	case TYPE_BUTTON: 
	case TYPE_EDITBOX:
	{
	  //printf("Token at line %i is Title\n", token_line );
	  stoken.ibuffer = strdup(s);
	  break;
	}
	default: {
          printf( "GRINGO: Token %s at line %i, pos. %i is not required !\n", s, token_line, token_count );
	  return 1;
	  break;
        }
      }
      break;
    }

    // Sixth
    case 6: {
      switch (stoken.type) {
        // Expect "Title"
	case TYPE_WINDOW: 
	{
	  //printf("Token at line %i is Title\n", token_line );
	  stoken.ibuffer = strdup(s);
	  break;
	}
	default: {
          printf( "GRINGO: Token %s at line %i, pos. %i is not required !\n", s, token_line, token_count );
	  return 1;
	  break;
        }
      }
      break;
    }
	
  //switch-token    
  }
  
  // Parse next token
  token_count++;
  return 0;
}




void token_end() {
  TObject *tempobj;
  int maxcount=0;
  if (stoken.type == TYPE_UNKNOWN) return;
  switch (stoken.type) {
    case TYPE_WINDOW: {
      maxcount=7;
      break;
    }      
    case TYPE_BUTTON:
    case TYPE_EDITBOX: {
      maxcount=6;
      break;
    }
    case TYPE_LABEL:
    {
      maxcount=5;
      break;
    }
  }
  if ( token_count != maxcount ) {
    printf( "GRINGO: Token at line %i is incomplete !\n", token_line );
    //TOKEN_FREE
    return;
  }
  
  // -----------------------------------------------

  if (stoken.type == TYPE_WINDOW ) {
    LastActiveWindow = (TObject*) window_Create( stoken.x, stoken.y, stoken.w, stoken.h, stoken.ibuffer );
    object_SetName( LastActiveWindow,  stoken.lname );
  }

  //TODO: Controllo esistenza finestra
  if ( LastActiveWindow == NULL ) {
    printf( "GRINGO: Devi definire prima una finestra, at line %i is incomplete !\n", token_line );
    gringo_Terminate(1);
    return;
  }

  if (stoken.type == TYPE_LABEL ) {
    label_Create( LastActiveWindow, stoken.y, stoken.x, stoken.w, stoken.ibuffer );
    // Labels has no names
  } else
  if (stoken.type == TYPE_BUTTON ) {
    tempobj = (TObject*)button_Create( LastActiveWindow, stoken.y, stoken.x, stoken.w, stoken.ibuffer );
    button_OnClick( tempobj, exit_from_gringosh );
    object_SetName( tempobj,  stoken.lname );
  } else
  if (stoken.type == TYPE_EDITBOX ) {
    tempobj = edit_Create( LastActiveWindow, stoken.y, stoken.x, stoken.w, stoken.ibuffer );
    object_SetName( tempobj,  stoken.lname );
  }
}



char* mystrtok( char *b, char s ) {
  int i,l;
  char *r;

  
  if ( b != NULL ) {
    strncpy( strtok_buffer, b, TOKBUFFER_SIZE-1 );
    strtok_buffer[TOKBUFFER_SIZE-1]=0;
    strtok_pointer = strtok_buffer;
  }

  r = NULL;
  l = strlen(strtok_pointer);
  i = 0;
  while ((i<l) && ( strtok_pointer[i] != s ) && ( strtok_pointer[i] != 0 ) ) i++;
  if (i>0) {
    r = (char*) malloc( sizeof(char) * (i+1) );
    strncpy( r, strtok_pointer, i );
    strtok_pointer = strtok_pointer + i + 1;
    r[i]=0;
  } else
  // In case we pass a NULL token ,i.e. no char between separators, we put a single space
  if (l>0) {
    r = (char*) malloc( sizeof(char) * 2 );
    strtok_pointer = strtok_pointer + 1;
    r[0]=32;
    r[1]=0;
  }

  return r;
}


void ParseInputFile(FILE *localfile)
 {
  char buffer[BUFFER_SIZE];
  char *token;
  const char separators   = ',';
  int count=0;
  
  //printf("Starting ...\n");
  while ( ! feof(localfile) ) {

    buffer[0]=0;
    count=0;
    while ( ! feof(localfile) ) {
      fread( &(buffer[count]), sizeof(char), 1, localfile );
      if ( ( buffer[count] == '\n' )&&(count<BUFFER_SIZE) ) {
        buffer[count] = 0;
	break;
      }
      count++;
    }

    //printf("\n?> %s\n", (char*)buffer );
    token = mystrtok( buffer, separators );
    token_begin();
    while( token != NULL )
 {
      /* While there are tokens in "string" */
      //printf( "!>%s\n", token );

      /* Parse Token */
      if ( token_parse(token) != 0 ) break;

      /* Get next token: */
      token = mystrtok( NULL, separators );
    }
    token_end();
  }
}

//EOF